import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import normal_init
from mmdet import models
from mmdet.ops import RoIPool, DeformConv
def conv2d_dw_group(x, kernel):
    batch, channel = kernel.shape[:2]
    x = x.view(1, batch*channel, x.size(2), x.size(3))  # 1 * (b*c) * k * k
    kernel = kernel.view(batch*channel, 1, kernel.size(2), kernel.size(3))  # (b*c) * 1 * H * W
    out = F.relu(F.conv2d(x, kernel, groups=batch*channel))
    out = out.view(batch, channel, out.size(2), out.size(3))
    return out

class DepthCorr(nn.Module):
    def __init__(self, in_channels, hidden, out_channels, kernel_size=3):
        super(DepthCorr, self).__init__()
        # adjust layer for asymmetrical features
        self.conv_kernel = nn.Sequential(
                nn.Conv2d(in_channels, hidden, kernel_size=kernel_size, bias=False),
                nn.BatchNorm2d(hidden),
                nn.ReLU(inplace=True),
                )
        self.conv_search = nn.Sequential(
                nn.Conv2d(in_channels, hidden, kernel_size=kernel_size, bias=False),
                nn.BatchNorm2d(hidden),
                nn.ReLU(inplace=True),
                )
        self.head = nn.Sequential(
                nn.Conv2d(hidden, hidden, kernel_size=1, bias=False),
                nn.BatchNorm2d(hidden),
                nn.ReLU(inplace=True),
                nn.Conv2d(hidden, out_channels, kernel_size=1)
                )

    def forward_corr(self, kernel, input):
        kernel = self.conv_kernel(kernel)
        input = self.conv_search(input)
        feature = conv2d_dw_group(input, kernel)
        return feature

    def forward(self, kernel, search):
        feature = self.forward_corr(kernel, search)
        out = self.head(feature)
        return out,feature
class DepthCorrDCN(nn.Module):
    def __init__(self, in_channels, hidden, out_channels, kernel_size=3):
        super(DepthCorrDCN, self).__init__()
        # adjust layer for asymmetrical features
        self.conv_offset= nn.Conv2d(2,18,1,bias=False)
        self.dcn = DeformConv(hidden,hidden,3,padding=1)
        self.head = nn.Sequential(
                nn.Conv2d(hidden, hidden, kernel_size=1, bias=False),
                nn.BatchNorm2d(hidden),
                nn.ReLU(inplace=True),
                nn.Conv2d(hidden, out_channels, kernel_size=1)
        )
        # normal_init(self.conv_offset,std=0.1)
        # normal_init(self.dcn,std=0.01)

    def forward(self, shape, feature):
        offset = self.conv_offset(shape)
        feature = F.relu(self.dcn(feature, offset))
        out = self.head(feature)
        return out
class ResNet(nn.Module):
    def __init__(self, init_weights=True):
        super(ResNet,self).__init__()
        cfg = dict(
                type='ResNet',
                depth=50,
                num_stages=3,
                strides=(1,2,1),
                dilations=(1,1,2),
                out_indices=(2,),
                frozen_stages=3,
                dcn=dict(
                    modulated=False,
                    deformable_groups=1,
                    fallback_on_stride=False
                    ),
                stage_with_dcn=(False,False,False)
                )
        self.model = models.build_backbone(cfg)
        self.model.conv1.padding=(0,0)
        self.model.maxpool.padding=(0,0)
        if init_weights:
            self.model.init_weights(pretrained='modelzoo://resnet50')
    def forward(self,x):
        return self.model(x)[0]
        
class SiamRPN(nn.Module):
    def __init__(self,feature_in=256, feature_out=256, anchor_num=3):
        super(SiamRPN, self).__init__()
        self.featureExtract = ResNet()
        self.downsample = nn.Sequential(
            nn.Conv2d(1024,256,1,bias=False),
            nn.BatchNorm2d(256),
        )
        self.cls = DepthCorrDCN(feature_in,feature_out,2*anchor_num)
        self.reg = DepthCorrDCN(feature_in,feature_out,4*anchor_num)
        self.loc = DepthCorr(feature_in,feature_out,1)
        self.shape = DepthCorr(feature_in,feature_out,2)
        torch.nn.init.constant_(self.loc.head[3].bias,-4.5911985)
        torch.nn.init.constant_(self.cls.head[3].bias,-4.5911985)


    def forward(self, x, z):
        z_f = self.featureExtract(z)
        z_f = self.downsample(z_f)
        z_f = z_f[:,:,4:-4,4:-4]
        x_f = self.featureExtract(x)
        x_f = self.downsample(x_f)
        loc,loc_feature = self.loc(z_f,x_f)
        shape,shape_feature = self.shape(z_f,x_f)
        cls = self.cls(shape.detach(),loc_feature)
        reg = self.reg(shape.detach(),shape_feature)
        shape = 8*8*torch.exp(shape)
        return reg,cls,loc,shape

class xSiamRPN(nn.Module):
    def __init__(self,feature_in=256, feature_out=256, anchor_num=3):
        super(xSiamRPN, self).__init__()
        self.featureExtract = ResNet(init_weights=False)
        self.cls = DepthCorrDCN(feature_in,feature_out,2*anchor_num)
        self.reg = DepthCorrDCN(feature_in,feature_out,4*anchor_num)
        self.loc = DepthCorr(feature_in,feature_out,1)
        self.shape = DepthCorr(feature_in,feature_out,2)
        self.downsample = nn.Sequential(
            nn.Conv2d(1024,256,1,bias=False),
            nn.BatchNorm2d(256),
        )

    def forward(self, x):
        x_f = self.featureExtract(x)
        x_f = self.downsample(x_f)
        loc,loc_feature = self.loc(self.z_f,x_f)
        shape,shape_feature = self.shape(self.z_f,x_f)
        cls = self.cls(shape,loc_feature)
        reg = self.reg(shape,shape_feature)
        shape = 8*8*torch.exp(shape)
        return reg,cls,loc,shape

    def temple(self,z):
        z_f = self.featureExtract(z)
        z_f = self.downsample(z_f)
        self.z_f = z_f[:,:,4:-4,4:-4]

