# --------------------------------------------------------
# DaSiamRPN
# Licensed under The MIT License
# Written by Qiang Wang (wangqiang2015 at ia.ac.cn)
# --------------------------------------------------------
import numpy as np
import cv2
import torch
from torch.autograd import Variable
import torch.nn.functional as F
from torchvision import transforms

from generate_anchor import generate_anchor
from utils import get_subwindow_tracking
from loc2bbox import loc2bbox


transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ])


def bayesian_SiamRPN_init(im, target_pos, target_sz, net, p):
    state = dict()
    state['im_h'] = im.shape[0]
    state['im_w'] = im.shape[1]
    p.instance_size = 55+8*p.score_size  # small object big search region
    
    avg_chans = np.mean(im, axis=(0, 1))
    wc_z = target_sz[0] + p.context_amount * sum(target_sz)
    hc_z = target_sz[1] + p.context_amount * sum(target_sz)
    s_z = round(np.sqrt(wc_z * hc_z))
    # initialize the exemplar
    im = get_subwindow_tracking(im, target_pos, p.exemplar_size, s_z, avg_chans)

    z = Variable(transform(im).unsqueeze(0))
    net.temple(z.cuda())

    shift_y = np.arange(0,p.score_size*p.total_stride,p.total_stride)
    shift_x = np.arange(0,p.score_size*p.total_stride,p.total_stride)
    shift_x,shift_y = np.meshgrid(shift_x,shift_y)
    shift = np.stack((
        shift_y.ravel(),
        shift_x.ravel(),
        shift_y.ravel(),
        shift_x.ravel()
    ),axis=1).reshape(-1,1,4)
    p.shift = shift+31

    state['p'] = p
    state['net'] = net
    state['avg_chans'] = avg_chans
    state['window'] = np.tile(gaussian_map(p.score_size,p.sigma)[:,:,np.newaxis],[1,1,3]).flatten()
    state['target_pos'] = target_pos
    state['target_sz'] = target_sz
    state['bayes_model'] = bayes_model_init(p.score_size)
    return state

def bayesian_SiamRPN_track(state, im):
    p = state['p']
    net = state['net']
    avg_chans = state['avg_chans']
    target_pos = state['target_pos']
    target_sz = state['target_sz']
    window = state['window']
    bayes_model = state['bayes_model']

    wc_z = target_sz[1] + p.context_amount * sum(target_sz)
    hc_z = target_sz[0] + p.context_amount * sum(target_sz)
    s_z = np.sqrt(wc_z * hc_z)
    scale_z = p.exemplar_size / s_z
    d_search = (p.instance_size - p.exemplar_size) / 2# 搜索区域大小
    pad = d_search / scale_z
    s_x = s_z + 2 * pad

    # extract scaled crops for search region x at previous target position
    x_crop_ori = get_subwindow_tracking(im, target_pos, p.instance_size, round(s_x), avg_chans)
    x_crop = Variable(transform(x_crop_ori).unsqueeze(0))
    target_pos, target_sz, score,bayes_model = bayes_tracker_eval(net, x_crop.cuda(), target_pos, target_sz * scale_z, scale_z, p, window,bayes_model)
    target_pos[0] = max(0, min(state['im_w'], target_pos[0]))
    target_pos[1] = max(0, min(state['im_h'], target_pos[1]))
    target_sz[0] = max(10, min(state['im_w'], target_sz[0]))
    target_sz[1] = max(10, min(state['im_h'], target_sz[1]))
    state['target_pos'] = target_pos
    state['target_sz'] = target_sz
    state['score'] = score
    state['bayes_model'] = bayes_model
    return state

### Bayesian anchoring

def gaussian_map(sz, sigma=5, shift=[0,0]):
    # when sigma=1.5x, Gau_map(x,x) equals to 0.8
    x = np.arange(-(sz-1)/2,(sz+1)/2)
    xx,yy   = np.meshgrid(x,x)
    Gau_map = np.exp(-0.5/(sigma**2)*((xx-shift[0])**2 + (yy-shift[1])**2))
    return Gau_map

def bayes_model_init(sz):
    bayes_model   = {}
    # bayes_model['sz']      = res_sz
    bayes_model['stride']  = 8
    bayes_model['shift']   = np.zeros([1,2]) #(x,y)
    bayes_model['accele']  = np.zeros([1,2]) #(x,y)
    # bayes_model['state']   = init_state
    bayes_model['seqs']    = [[0,0,0,0]]
    bayes_model['Gau_map'] = np.tile(gaussian_map(sz)[:,:,np.newaxis],[1,1,3]).flatten()#gaussian_map(res_sz)
    # bayes_model['trans_state'] = np.tile(init_state[:,:,np.newaxis],[1,1,3]).flatten()
    return bayes_model

def bayes_inference(bayes_model, new_observ, cur_state,sz):
# this function computes the bayesian model with the new observation
    seqs      = bayes_model['seqs']
    stride    = bayes_model['stride']

    seqs.append(new_observ)
    if len(seqs)<5:
        bayes_model['seqs'] = seqs
        return bayes_model
    # update the transition model when the recorded states are more than 4
    seqs_array = np.array(seqs,dtype=float)
    shift_1    = (seqs_array[-3:,0:2]*np.array([[0.2], [0.3], [0.5]])).sum(0)/stride # compute the average value of recent speeds as current speed
    accele     = ((seqs_array[-4:,0:2]-seqs_array[-5:-1,0:2])*np.array([[0.1], [0.2], [0.3], [0.4]])).sum(0)
    shift_2    = accele/stride
    shift      = shift_1 + shift_2
    shift = np.clip(shift,-1,1)
    # print('shift:',shift)
    bayes_model['Gau_map'] = np.tile(gaussian_map(sz=sz,shift=shift)[:,:,np.newaxis],[1,1,3]).flatten()#gaussian_map(res_sz)
    seqs.pop(0)
    return bayes_model


def bayes_tracker_eval(net, x_crop, target_pos, target_sz, scale_z, p, window,bayes_model):
    delta, score, loc, shape = net(x_crop)

    delta = delta.permute(0, 2, 3, 1).reshape(-1, 4).cpu().detach().numpy()
    score = F.softmax(score.permute(0, 2, 3, 1).reshape(-1, 2), dim=1).cpu().detach().numpy()[:, 1]
    loc = loc.permute(0, 2, 3, 1).reshape(-1).sigmoid().cpu().detach().numpy()
    hw = shape.permute(0, 2, 3, 1).reshape(-1, 2).cpu().detach().numpy()
    hw = np.stack((
        hw[:, 0] * np.sqrt(3 / 2),
        hw[:, 1] * np.sqrt(2 / 3),
        hw[:, 0],
        hw[:, 1],
        hw[:, 0] * np.sqrt(2 / 3),
        hw[:, 1] * np.sqrt(3 / 2),
    ), axis=-1).reshape(p.score_size * p.score_size, -1, 2)
    anchor = np.stack((
        p.shift[:, :, 0] - hw[:, :, 0] / 2,
        p.shift[:, :, 1] - hw[:, :, 1] / 2,
        p.shift[:, :, 2] + hw[:, :, 0] / 2,
        p.shift[:, :, 3] + hw[:, :, 1] / 2,
    ), axis=-1)
    anchor = anchor.reshape(-1, 4)
    delta = loc2bbox(anchor, delta).transpose()
    delta[:2] -= p.instance_size/2

    def change(r):
        return np.maximum(r, 1. / r)
    def sz(w, h):
        pad = (w + h) * 0.5
        sz2 = (w + pad) * (h + pad)
        return np.sqrt(sz2)
    def sz_wh(wh):
        pad = (wh[0] + wh[1]) * 0.5
        sz2 = (wh[0] + pad) * (wh[1] + pad)
        return np.sqrt(sz2)

    # size penalty
    s_c = change(sz(delta[2, :], delta[3, :]) / (sz_wh(target_sz)))  # scale penalty
    r_c = change((target_sz[0] / target_sz[1]) / (delta[2, :] / delta[3, :]))  # ratio penalty
    penalty = np.exp(-(r_c * s_c - 1.) * p.penalty_k)
    pscore = penalty * score

    # pscore = pscore * (1 - p.window_influence) + window * p.window_influence
    pscore = pscore * (1 - p.window_influence) + bayes_model['Gau_map']* p.window_influence
    best_pscore_id = np.argmax(pscore)

    target = delta[:, best_pscore_id] / scale_z
    target_sz = target_sz / scale_z
    lr = penalty[best_pscore_id] * score[best_pscore_id] * p.lr
    bayes_model = bayes_inference(bayes_model, delta[:,best_pscore_id], pscore.reshape(p.score_size,p.score_size,-1).max(2),p.score_size) #delta--(delta_x, delta_y, w, h)
    res_x = target[0] + target_pos[0]
    res_y = target[1] + target_pos[1]
    res_w = target_sz[0] * (1 - lr) + target[2] * lr
    res_h = target_sz[1] * (1 - lr) + target[3] * lr
    target_pos = np.array([res_x, res_y])
    target_sz = np.array([res_w, res_h])
    return target_pos, target_sz, score[best_pscore_id],bayes_model
